# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/google/GDrive.ipynb.

# %% auto 0
__all__ = ['get_file_by_id', 'get_folder_by_id', 'walk_folder', 'GDrive_MimeType_Enum', 'ExportAs_Enum', 'GDrive_ExportError',
           'export_file_as', 'GDrive_ServiceRequired', 'GDrive_File', 'GDrive_Files']

# %% ../../nbs/google/GDrive.ipynb 2
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Any, Optional

import datetime as dt


from googleapiclient.errors import HttpError

from domolibrary_extensions.utils.utils import (
    upsert_folder,
    convert_str_to_date,
    convert_str_file_name,
    download_zip,
    download_pptx,
)

import domolibrary_extensions.google.GAuth as ga

from nbdev.showdoc import patch_to

# %% ../../nbs/google/GDrive.ipynb 5
def get_file_by_id(service, file_id, fields: List[str] = None, parent_ls=None):
    parent_ls = parent_ls or []

    fields = fields or [
        "id",
        "webViewLink",
        "name",
        "mimeType",
        " modifiedTime",
        "parents",
    ]

    file = service.files().get(fileId=file_id, fields=", ".join(fields)).execute()

    parent_ls += [
        parent for parent in file.get("parents", []) if parent not in parent_ls
    ]

    file.update({"parent_ls": parent_ls})
    return file

# %% ../../nbs/google/GDrive.ipynb 6
def get_folder_by_id(
    service,
    folder_id,
    parent_ls: Optional[
        List[str]
    ] = None,  # used by recursive query to build path to file
):
    parent_ls = parent_ls or []
    file_ls = []
    page_token = None

    while True:
        res = (
            service.files()
            .list(
                q=f"'{folder_id}' in parents",
                pageSize=10,
                fields="nextPageToken, files(id,webViewLink, name, mimeType,modifiedTime, parents )",
                pageToken=page_token,
            )
            .execute()
        )

        file_ls.extend(res.get("files", []))

        page_token = res.get("nextPageToken", None)
        if not page_token:
            break

    for file in file_ls:
        parent_ls += [
            parent for parent in file.get("parents", []) if parent not in parent_ls
        ]
        file.update({"parent_ls": parent_ls})

    return file_ls

# %% ../../nbs/google/GDrive.ipynb 7
def walk_folder(
    service,
    folder_id: str,
    file_ls: List[dict],
    is_recursive: bool = True,
    parent_ls: Optional[List[str]] = None,
):  # recursive function that retrieves all files in subfolders

    parent_ls = parent_ls or [folder_id]
    new_files = get_folder_by_id(service, folder_id, parent_ls=parent_ls)
    file_ls.extend(new_files)

    if new_files and is_recursive:
        [
            walk_folder(
                service=service,
                folder_id=file_obj["id"],
                file_ls=file_ls,
                parent_ls=[
                    *parent_ls,
                    *[
                        parent
                        for parent in file_obj["parent_ls"]
                        if parent not in parent_ls
                    ],
                ],
            )
            for file_obj in new_files
            if file_obj.get("mimeType") == "application/vnd.google-apps.folder"
        ]

# %% ../../nbs/google/GDrive.ipynb 9
class GDrive_MimeType_Enum(Enum):
    """for translating Google Drive types into 'exports as'"""

    docx = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    pptx = "application/vnd.openxmlformats-officedocument.presentationml.presentation"

    odt = "application/vnd.oasis.opendocument.text"
    rtf = "application/rtf"
    pdf = "application/pdf"
    txt = "text/plain"

    zip_file = "application/zip"
    epub = "application/epub+zip"

    folder = "application/vnd.google-apps.folder"
    gdoc = "application/vnd.google-apps.document"
    gsheet = "application/vnd.google-apps.spreadsheet"
    gshortcut = "application/vnd.google-apps.shortcut"
    gslides = "application/vnd.google-apps.presentation"

    # html = 'application/zip'
    # 'tex'  = 'application/zip'
    # 'html.zip'= 'application/zip'


class ExportAs_Enum(Enum):
    pptx = {
        "download_fn": download_pptx,
        "export_mime_type": GDrive_MimeType_Enum.pptx,
    }
    default = {
        "download_fn": download_zip,
        "export_mime_type": GDrive_MimeType_Enum.zip_file,
    }

    @classmethod
    def _missing_(
        cls, value
    ):  # if export as not defined, will return default of export as zip
        return cls.default

# %% ../../nbs/google/GDrive.ipynb 10
class GDrive_ExportError(Exception):
    def __init__(self, doc_url):
        message = f"failure to download content for {doc_url}"
        super().__init__(message)


def export_file_as(service, file_id, export_as: ExportAs_Enum, output_path):

    export_mime_type, download_fn = (
        export_as.value["export_mime_type"].value,
        export_as.value["download_fn"],
    )

    loop = 0
    content = None
    while loop < 3 and content is None:
        loop += 1
        try:
            content = (
                service.files()
                .export(fileId=file_id, mimeType=export_mime_type)
                .execute()
            )

        except HttpError as err:
            print(err)

        except Exception as e:
            print(e)

    if not content:
        raise GDrive_ExportError(file_id)

    upsert_folder(output_path)

    return download_fn(content, output_path)

# %% ../../nbs/google/GDrive.ipynb 13
class GDrive_ServiceRequired(Exception):
    def __init__(self, doc_url=None, doc_id=None):
        message = f"service (googleclientapi.discovery.Resource) required to download {doc_url or doc_id}"
        super().__init__(message)


@dataclass
class GDrive_File:
    id: str
    name: str

    modified_time: dt.datetime
    url: str

    auth: ga.GoogleAuth = field(repr=False)
    service: Any = field(repr=False)

    mime_str: str

    parent_ls: List[str] = field(default=None)

    gdrive_mime_type: GDrive_MimeType_Enum = None
    gdrive_parents: List[str] = field(default=None, repr=False)

    file: Any = field(repr=False, default=None)

    def __post_init__(self):
        try:
            self.gdrive_mime_type = GDrive_MimeType_Enum(self.mime_str)

        except Exception as e:
            print(e)

        if self.parent_ls:
            self.gdrive_parents = [
                GDrive_File.get_by_id(
                    document_id=parent, auth=self.auth, service=self.service
                )
                for parent in self.parent_ls
            ]

    @classmethod
    def _from_json(cls, obj: dict, auth: ga.GoogleAuth, service: None):
        return cls(
            auth=auth,
            service=service,
            file=obj,
            mime_str=obj["mimeType"],
            url=str(obj["webViewLink"]),
            id=obj["id"],
            name=obj["name"],
            parent_ls=obj["parent_ls"],
            modified_time=convert_str_to_date(obj["modifiedTime"]),
        )

    @classmethod
    def get_by_id(
        cls,
        document_id: str,
        auth: ga.GoogleAuth = None,
        service=None,
        return_raw: bool = False,
    ):
        service = service or auth.generate_service("drive", "v3")

        file = get_file_by_id(file_id=document_id, service=service)

        if return_raw:
            return file

        return cls._from_json(obj=file, auth=auth, service=service)

    def export(self, output_path, export_as: ExportAs_Enum = ExportAs_Enum.default):

        return export_file_as(
            service=self.service,
            file_id=self.id,
            export_as=export_as,
            output_path=output_path,
        )

# %% ../../nbs/google/GDrive.ipynb 18
@dataclass
class GDrive_Files:

    auth: ga.GoogleAuth = field(repr=False)

    folder_id: str = None

    gdrive_files: List[GDrive_File] = field(default_factory=lambda: [])

    service: ga.Resource = field(default=None, repr=False)

    def __post_init__(self):

        self.service = self.service or self.auth.generate_service(
            service_name="drive", service_version="v3"
        )

    def get_files(self, folder_id, is_recursive: bool = True, return_raw: bool = False):
        file_ls = []

        walk_folder(
            service=self.service,
            folder_id=folder_id,
            file_ls=file_ls,
            is_recursive=is_recursive,
        )

        if return_raw:
            return file_ls

        self.gdrive_files = [
            GDrive_File._from_json(obj=obj, auth=self.auth, service=self.service)
            for obj in file_ls
        ]

        return self.gdrive_files



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/jira/classes.ipynb.

# %% ../../nbs/jira/classes.ipynb 2
from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

import datetime as dt
import httpx

from nbdev.showdoc import patch_to

import domolibrary_extensions.jira.auth as ja
import domolibrary_extensions.jira.routes as dejr
import domolibrary_extensions.utils.convert as dxcv

# %% auto 0
__all__ = ['JiraPerson', 'JiraProject', 'JiraPriority', 'JiraComment', 'JiraIssueType', 'JiraRemoteLink', 'JiraIssue']

# %% ../../nbs/jira/classes.ipynb 3
@dataclass
class JiraPerson:
    url: str
    display_name: str
    email_address: str
    avatar_url: str

    @classmethod
    def from_json(cls, obj):
        return cls(
            url=obj["self"],
            display_name=obj["displayName"],
            email_address=obj["emailAddress"],
            avatar_url=obj["avatarUrls"]["48x48"],
        )


@dataclass
class JiraProject:
    uid: str
    id: str
    name: str
    url: str

    @classmethod
    def from_json(cls, obj):
        return cls(uid=obj["id"], id=obj["key"], name=obj["name"], url=obj["self"])


@dataclass
class JiraPriority:
    id: str
    url: str
    title: str

    @classmethod
    def from_json(cls, obj):
        return cls(id=obj["id"], url=obj["self"], title=obj["name"])

# %% ../../nbs/jira/classes.ipynb 4
@dataclass
class JiraComment:
    auth: ja.JiraAuth = field(repr=False)
    jira_author: JiraPerson
    body: str
    created_dt: dt.datetime
    updated_dt: dt.datetime
    rendered_body: str = None

    @classmethod
    def from_json(cls, obj, auth):
        return cls(
            auth=auth,
            jira_author=JiraPerson.from_json(obj["author"]),
            body=obj["body"],
            rendered_body=obj.get("renderedBody"),
            created_dt=dxcv.convert_str_to_date(obj["created"]),
            updated_dt=dxcv.convert_str_to_date(obj["updated"]),
        )

# %% ../../nbs/jira/classes.ipynb 5
@dataclass
class JiraIssueType:
    id: str
    url: str
    description: str

    @classmethod
    def from_json(cls, obj):
        return cls(id=obj["name"], url=obj["self"], description=obj["description"])


@dataclass
class JiraRemoteLink:
    auth: ja.JiraAuth = field(repr=False)
    url: str  # object.url
    title: str  # object.title

    id: str = None  # internal id
    internal_url: str = None  # self
    summary: str = None  # object.summary

    relationship: str = "links to"  # relationship

    global_id: str = None  # globalId, how user identifies entity
    icon_url: str = None  # object.icon.url16x16

    resolved: bool = None  # object.status.resolved
    application_name: str = None  # application.name

    @classmethod
    def from_json(cls, obj, auth):

        return JiraRemoteLink(
            auth=auth,
            id=obj.get("id"),
            internal_url=obj.get("self"),
            global_id=obj.get("globalId"),
            url=obj["object"]["url"],
            title=obj["object"]["title"],
            summary=obj["object"].get("summary"),
            relationship=obj.get("relationship"),
            icon_url=obj["object"].get("icon", {}).get("url16x16"),
            resolved=obj["object"].get("status", {}).get("resolved"),
            application_name=obj["application"].get("name"),
        )

    def to_json(self):
        obj = {
            "globalId": self.global_id,
            "id": self.id,
            "application": {"name": self.application_name},
            "relationship": self.relationship,
            "object": {
                "url": self.url,
                "title": self.title,
                "summary": self.summary,
                "icon": {"url16x16": self.icon_url},
                "status": {"resolved": self.resolved},
            },
            "self": self.internal_url,
        }

        return obj


@dataclass
class JiraIssue:
    auth: ja.JiraAuth = field(repr=False)
    uid: str
    id: str
    title: str
    url: str

    created_dt: dt.datetime
    updated_dt: dt.datetime
    duedate_dt: dt.datetime

    labels: List[str]
    description: str

    jira_status: dict
    # jira_status_category : dict
    jira_issue_type: dict
    jira_priority: JiraPriority
    jira_reporter: JiraPerson
    jira_project: dict

    jira_comments: List[JiraComment]

    jira_links: List[JiraRemoteLink] = None

    @classmethod
    def from_json(cls, obj, auth):

        fields = obj.get("fields")

        return cls(
            auth=auth,
            uid=obj["id"],
            id=obj["key"],
            url=obj["self"],
            title=fields["summary"],
            created_dt=dxcv.convert_str_to_date(fields["created"]),
            duedate_dt=dxcv.convert_str_to_date(fields["duedate"]),
            updated_dt=dxcv.convert_str_to_date(fields["updated"]),
            labels=fields["labels"],
            description=fields["description"],
            jira_status=fields["status"],
            # jira_status_category = fields['statusCategory'],
            jira_issue_type=JiraIssueType.from_json(fields["issuetype"]),
            jira_project=JiraProject.from_json(fields["project"]),
            jira_priority=JiraPriority.from_json(fields["priority"]),
            jira_reporter=JiraPerson.from_json(fields["reporter"]),
            jira_comments=[
                JiraComment.from_json(obj=comment, auth=auth)
                for comment in fields.get("comment", {}).get("comments", [])
            ],
        )

# %% ../../nbs/jira/classes.ipynb 6
@patch_to(JiraIssue, cls_method=True)
async def get_by_id(
    cls,
    issue_id: int,
    auth: ja.JiraAuth,
    debug_api: bool = False,
    return_raw: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):

    res = await dejr.get_issue_by_id(
        auth=auth,
        issue_id=issue_id,
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
    )

    if return_raw:
        return res

    return cls.from_json(obj=res.response, auth=auth)


@patch_to(JiraIssue)
async def get_comments(
    self,
    debug_api: bool = False,
    return_raw: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):

    res = await dejr.get_issue_comments(
        issue_id=self.id,
        auth=self.auth,
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
    )

    if return_raw:
        return res

    self.jira_comments = [
        JiraComment.from_json(obj=comment, auth=self.auth) for comment in res.response
    ]

    return self.jira_comments

# %% ../../nbs/jira/classes.ipynb 7
@patch_to(JiraIssue)
async def get_remotelinks(
    self,
    debug_api: bool = False,
    return_raw: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):
    res = await dejr.get_issue_remotelinks(
        auth=self.auth,
        issue_id=self.id,
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
    )

    if return_raw:
        return res

    self.jira_links = [
        JiraRemoteLink.from_json(link, auth=self.auth) for link in res.response
    ]

    return self.jira_links

# %% ../../nbs/jira/classes.ipynb 8
@patch_to(JiraIssue)
async def upsert_remotelink(
    self,
    remote_link: JiraRemoteLink,
    debug_api: bool = False,
    return_raw: bool = False,
    is_ignore_cache: bool = False,
    cache_path: str = None,
    client: httpx.AsyncClient = None,
    timeout: int = 30,
):
    res = await dejr.upsert_issue_remotelink(
        auth=self.auth,
        issue_id=self.id,
        body=remote_link.to_json(),
        debug_api=debug_api,
        is_ignore_cache=is_ignore_cache,
        cache_path=cache_path,
        client=client,
        timeout=timeout,
    )

    if return_raw:
        return res

    return await self.get_remotelinks()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/1pass/op.ipynb.

# %% auto 0
__all__ = ['PreferredItemCategory', 'KnownItemType', 'KnownItemPurpose', 'generate_password_recipe', 'Vault', 'Item_Section',
           'Item_Field', 'Vault_Item']

# %% ../nbs/1pass/op.ipynb 3
from onepasswordconnectsdk.async_client import AsyncClient as Client

# %% ../nbs/1pass/op.ipynb 4
from dataclasses import dataclass, field
from typing import Any, List, Union
from enum import Enum, auto

import datetime as dt
from pprint import pprint

from onepasswordconnectsdk.models import Item, Field
from onepasswordconnectsdk.models.summary_item import SummaryItem

from domolibrary_extensions.utils import utils as deu
import domolibrary_extensions.utils.execution as dex
import domolibrary_extensions.client as cl

from nbdev.showdoc import patch_to

# %% ../nbs/1pass/op.ipynb 8
class PreferredItemCategory(Enum):
    """preferred types
    complete list of types can be retireved here https://developer.1password.com/docs/connect/connect-api-reference/#add-an-item
    """

    LOGIN = auto()
    PASSWORD = auto()
    API_CREDENTIAL = auto()
    SSH_KEY = auto

# %% ../nbs/1pass/op.ipynb 10
class KnownItemType(Enum):
    """based on list from current vault"""

    STRING = auto()
    CONCEALED = auto()
    DATE = auto()
    MENU = auto()
    URL = auto()


class KnownItemPurpose(Enum):
    NOTES = auto()
    PASSWORD = auto()
    USERNAME = auto()


# set([field.type for item in vault.items for field in item.fields])
# KnownItemPurpose.NOTES.__class__.__name__

# %% ../nbs/1pass/op.ipynb 11
def clean_dict(obj):
    return {key: value for key, value in obj.items() if value}

# %% ../nbs/1pass/op.ipynb 12
def generate_password_recipe(
    length: int = 24, is_letters: bool = True, is_digits: bool = True
):
    characterSets = []
    if is_letters:
        characterSets.append("LETTERS")

    if is_digits:
        characterSets.append("DIGITS")

    return {"length": length, "characterSes": characterSets}

# %% ../nbs/1pass/op.ipynb 14
@dataclass
class Vault:
    client: Client = field(repr=False)  # pass an asynchronous client
    id: str
    name: str

    vault_items = None

    @classmethod
    def from_sdk(cls, obj, client: Client):
        return cls(id=obj.id, name=obj.name, client=client)

    @classmethod
    async def get_vault(
        cls,
        aclient: Client,
        vault_name: str = None,
        vault_id: str = None,
        return_raw: bool = False,
    ):  # pass an asyc enabled client
        sdk_obj = None

        if vault_id:
            sdk_obj = await aclient.get_vault(vault_id)

        if vault_name:
            sdk_obj = await aclient.get_vault_by_title(vault_name)

        if return_raw:
            return sdk_obj

        return cls.from_sdk(sdk_obj, aclient)

# %% ../nbs/1pass/op.ipynb 16
@dataclass
class Item_Section:
    id: str
    label: str = None

    """
    used to group fields in an item.
    must put custom fields into a section
    """

    @classmethod
    def _from_sdk(cls, sdk_obj):
        return cls(id=sdk_obj.id, label=sdk_obj.label)

    def to_sdk(self):
        return clean_dict({"id": self.id, "label": self.label})

    def to_sdk_for_item(self):
        return clean_dict({"id": self.id})

# %% ../nbs/1pass/op.ipynb 18
@dataclass
class Item_Field:
    label: str = None
    value: str = None
    recipe: dict = None
    id: str = None
    is_call_generate_fn: bool = False
    is_secret: bool = False

    client: Client = field(repr=False, default=None)
    type: Union[str, KnownItemType] = None
    purpose: Union[str, KnownItemPurpose] = ""
    item_section: Item_Section = None

    def __eq__(self, other):
        if not isinstance(other, Item_Field):
            return False

        if self.label == other.label:
            return True

        if self.id == other.id and (self.id is not None or other.id is not None):
            return True

        return False

    @classmethod
    def _from_sdk(cls, sdk_obj, client):
        item_section = Item_Section(sdk_obj.section)

        return cls(
            client=client,
            id=sdk_obj.id,
            item_section=item_section,
            label=sdk_obj.label,
            purpose=sdk_obj.purpose,
            value=sdk_obj.value,
            # recipe=sdk_obj.recipe,
            is_secret=True if sdk_obj.type == "CONCEALED" else False,
            type=sdk_obj.type,
            is_call_generate_fn=sdk_obj.generate,
        )

    def generate_recipie(
        self, length: int = 24, is_letters: bool = True, is_digits: bool = True
    ):
        """generates a recipie for how generating passwords is handled"""

        self.recipe = generate_password_recipe(
            length=length, is_letters=is_letters, is_digits=is_digits
        )
        return self.recipe

    def to_sdk(self):
        return Field(
            id=self.id,
            label=self.label,
            value=self.value,
            purpose=deu.convert_enum_member_to_name(self.purpose),
            type=deu.convert_enum_member_to_name(self.type),
            generate=self.is_call_generate_fn,
            # recipe = self.recipe,
            section=self.item_section.to_sdk_for_item() if self.item_section else None,
        )

    def __repr__(self):
        kws = [
            f"{key}={value!r}"
            for key, value in self.__dict__.items()
            if key not in ["client", "is_secret", "is_call_generate_fn"]
            and not (key == "value" and self.is_secret)
        ]
        return "{}({})".format(type(self).__name__, ", ".join(kws))

# %% ../nbs/1pass/op.ipynb 20
@dataclass
class Vault_Item:
    vault: Vault
    title: str
    category: Union[str, PreferredItemCategory]
    vault_fields: List[Item_Field]
    vault_sections: List[Item_Section] = None
    tags: List[str] = None
    created_at: dt.datetime = None
    updated_at: dt.datetime = None
    is_trashed: bool = False
    id: str = None
    client: Client = field(repr=False, default=None)

    def __eq__(self, other):
        if not isinstance(other, Vault_Item):
            return False

        if not self.vault.id == other.vault.id:
            return False

        if self.title == other.title:
            return True

        if self.id == other.id and (self.id is not None or other.id is not None):
            return True

        return False

    @classmethod
    async def _from_sdk(cls, sdk_obj, client):
        if isinstance(sdk_obj, SummaryItem):
            sdk_obj = await client.get_item_by_id(
                vault_id=sdk_obj.vault.id, item_id=sdk_obj.id
            )

        vault_sections = None
        if sdk_obj.sections:
            vault_sections = [
                Item_Section._from_sdk(section) for section in sdk_obj.sections
            ]

        vault_fields = [
            Item_Field._from_sdk(sdk_field, client=client)
            for sdk_field in sdk_obj.fields
        ]

        return cls(
            client=client,
            id=sdk_obj.id,
            vault=await Vault.get_vault(vault_id=sdk_obj.vault.id, aclient=client),
            vault_sections=vault_sections,
            title=sdk_obj.title,
            tags=sdk_obj.tags,
            category=sdk_obj.category,
            created_at=sdk_obj.created_at,
            updated_at=sdk_obj.updated_at,
            is_trashed=sdk_obj.trashed,
            vault_fields=vault_fields,
        )

    def to_sdk(self):
        return Item(
            # "vault": {"id": self.vault_id},
            id=self.id,
            title=self.title,
            category=deu.convert_enum_member_to_name(self.category),
            tags=self.tags,
            fields=(
                [vault_field.to_sdk() for vault_field in self.vault_fields]
                if self.vault_fields
                else None
            ),
            sections=(
                [vault_section.to_sdk() for vault_section in self.vault_sections]
                if self.vault_sections
                else None
            ),
        )

    @classmethod
    async def get_by_id(
        cls, vault_id, item_id, client: Client, return_raw: bool = False
    ):
        sdk_obj = await client.get_item_by_id(vault_id=vault_id, item_id=item_id)

        if return_raw:
            return sdk_obj

        return await cls._from_sdk(sdk_obj=sdk_obj, client=client)

# %% ../nbs/1pass/op.ipynb 22
@patch_to(Vault_Item)
async def delete(self: Vault_Item):
    await self.client.delete_item(self.id, self.vault.id)
    return True


@patch_to(Vault_Item)
async def update(self: Vault_Item, return_raw: bool = False, debug_api: bool = False):
    item_uuid = self.id
    vault_id = self.vault.id
    sdk_item = self.to_sdk()

    if debug_api:
        print({"item_uuid": item_uuid, "vaul_id": vault_id, "item": sdk_item})

    sdk_obj = await self.client.update_item(
        item_uuid=item_uuid, vault_id=vault_id, item=sdk_item
    )

    if return_raw:
        return sdk_obj

    return await Vault_Item._from_sdk(sdk_obj, client=self.client)


@patch_to(Vault_Item)
def add_field(self, vault_field: Item_Field):
    self.vault_fields = self.vault_fields or []

    if vault_field in self.vault_fields:
        target_index = self.vault_fields.index(vault_field)

        ele = self.vault_fields[target_index]
        vault_field.id = ele.id

        self.vault_fields[target_index] = vault_field

    else:
        self.vault_fields.append(vault_field)

    return self.vault_fields

# %% ../nbs/1pass/op.ipynb 26
class VaultItem_GET_Excception(cl.BaseError):
    def __init__(self, vault_str: str, message):  # id or name
        super().__init__(instance=vault_str, message=message)


@patch_to(Vault)
async def get_items(self: Vault, return_raw: bool = False) -> List[Vault_Item]:
    sdk_items = await self.client.get_items(self.id)

    if return_raw:
        return await dex.gather_with_concurrency(
            *[
                self.client.get_item_by_id(vault_id=self.id, item_id=sdk_item.id)
                for sdk_item in sdk_items
            ],
            n=10,
        )

    items = await dex.gather_with_concurrency(
        *[
            Vault_Item._from_sdk(sdk_obj=sdk_item, client=self.client)
            for sdk_item in sdk_items
        ],
        n=10,
    )

    self.items = items
    return items


@patch_to(Vault)
async def get_item(
    self: Vault, item_title=None, item_id=None, return_raw: bool = False
) -> Vault_Item:
    "retrieve an item by title or id"

    vault_item = None

    if item_id:
        vault_item = await self.client.get_item_by_id(
            vault_id=self.id, item_id=item_id, return_raw=return_raw
        )

    if not vault_item and item_title:
        vault_items = await self.get_items(return_raw=return_raw)

        vault_item = next(
            (
                vault_item
                for vault_item in vault_items
                if vault_item.title == item_title
            ),
            None,
        )

    if not vault_item:
        criteria = ""
        if item_id:
            criteria += item_id

        if item_title:
            criteria += f" - {item_title}"

        raise VaultItem_GET_Excception(
            vault_str=(self.name or self.id),
            message=f"no items returned with {criteria}",
        )

    return vault_item

# %% ../nbs/1pass/op.ipynb 28
@patch_to(Vault)
async def create_item(
    self: Vault,
    title: str,
    category: PreferredItemCategory,
    tags: List[str],
    vault_fields: List[Item_Field] = None,
    debug_api: bool = False,
    return_raw: bool = False,
):
    sdk_item = Item(
        title=title,
        category=deu.convert_enum_member_to_name(category),
        tags=tags,
        fields=(
            [vault_field.to_sdk() for vault_field in vault_fields]
            if vault_fields
            else None
        ),
    )

    if debug_api:
        pprint(sdk_item)

    sdk_item = await self.client.create_item(item=sdk_item, vault_id=self.id)

    if return_raw:
        return sdk_item

    return await Vault_Item._from_sdk(sdk_item, self.client)

# %% ../nbs/1pass/op.ipynb 29
@patch_to(Vault)
async def upsert_item(
    self: Vault,
    item_id: str = None,
    item_title: str = None,
    category: PreferredItemCategory = None,
    tags: List[str] = None,
    vault_fields: List[Item_Field] = None,
    debug_prn: bool = False,
    debug_api: bool = False,
):
    vault_fields = vault_fields or []

    vault_item = None

    # try to retrieve vault_item
    try:
        vault_item = await self.get_item(
            item_id=item_id,
            item_title=item_title,
        )

        # update item
        if item_title:
            vault_item.title = item_title

        if tags:
            vault_item.tags = tags

        for vault_field in vault_fields:
            vault_item.add_field(vault_field)

        if debug_prn:
            print("updating item")

        await vault_item.update(debug_api=debug_api)

    # create vault_item
    except VaultItem_GET_Excception as e:
        if debug_prn:
            print("creating item")

        vault_item = await self.create_item(
            title=item_title,
            category=category,
            tags=tags,
            vault_fields=vault_fields,
            debug_api=debug_api,
        )

    return vault_item



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/google/GDoc.ipynb.

# %% auto 0
__all__ = ['get_document_by_id', 'GDoc_StyleEnum', 'GDoc_Text_Style', 'GDoc_Element_TextRun', 'GDoc_Element',
           'GDoc_Paragraph_Style', 'GDoc_Paragraph', 'GDoc_SectionBreak_Style', 'GDoc_SectionBreak', 'GDoc_Content',
           'GDoc', 'DocState', 'generate_delete_json', 'generate_insert_json', 'generate_update_style',
           'generate_insert_hyperlink']

# %% ../../nbs/google/GDoc.ipynb 2
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Any, Optional, Callable


import domolibrary_extensions.google.GAuth as ga

# %% ../../nbs/google/GDoc.ipynb 5
def get_document_by_id(service, document_id):
    document = service.documents().get(documentId=document_id).execute()
    return document

# %% ../../nbs/google/GDoc.ipynb 7
class GDoc_StyleEnum(Enum):
    HEADING_1 = "HEADING_1"
    HEADING_2 = "HEADING_2"

# %% ../../nbs/google/GDoc.ipynb 8
@dataclass
class GDoc_Text_Style:
    pass


@dataclass
class GDoc_Element_TextRun:
    content: str

    textStyle: GDoc_Text_Style
    text: str

    @classmethod
    def _from_json(cls, obj):

        return cls(
            content=obj.get("content"),
            textStyle=obj.get("textStyle"),
            text=obj.get("content").strip(),
        )


@dataclass
class GDoc_Element:
    startIndex: int
    endIndex: int
    textRun: GDoc_Element_TextRun

    @classmethod
    def _from_json(cls, obj):
        return GDoc_Element(
            startIndex=obj.get("startIndex"),
            endIndex=obj.get("endIndex"),
            textRun=GDoc_Element_TextRun._from_json(obj.get("textRun")),
        )


@dataclass
class GDoc_Paragraph_Style:

    headingId: Optional[str]

    namedStyleType: str
    direction: str

    @classmethod
    def _from_json(cls, obj):

        return cls(
            headingId=obj.get("headingId"),
            namedStyleType=obj["namedStyleType"],
            direction=obj["direction"],
        )


@dataclass
class GDoc_Paragraph:

    elements: List[GDoc_Element]

    paragraphStyle: Optional[GDoc_Paragraph_Style]

    @classmethod
    def _from_json(cls, obj):

        return cls(
            elements=[
                GDoc_Element._from_json(element) for element in obj.get("elements")
            ],
            paragraphStyle=GDoc_Paragraph_Style._from_json(obj.get("paragraphStyle")),
        )


@dataclass
class GDoc_SectionBreak_Style:

    columnSeparatorStyle: str
    contentDirection: str

    sectionType: str


@dataclass
class GDoc_SectionBreak:

    sectionStyle: GDoc_SectionBreak_Style


@dataclass
class GDoc_Content:

    startIndex: Optional[int]

    endIndex: Optional[int]

    paragraph: Optional[GDoc_Paragraph]

    sectionBreak: Optional[GDoc_SectionBreak]

    @classmethod
    def _from_json(cls, obj):

        gdoc_paragraph = (
            GDoc_Paragraph._from_json(obj.get("paragraph"))
            if obj.get("paragraph")
            else None
        )

        return cls(
            startIndex=obj.get("startIndex"),
            endIndex=obj.get("endIndex"),
            sectionBreak=obj.get("sectionBreak"),
            paragraph=gdoc_paragraph,
        )

# %% ../../nbs/google/GDoc.ipynb 9
@dataclass
class GDoc:
    auth: ga.GoogleAuth = field(repr=False)
    service: Any = field(repr=False)
    document_id: str
    # url: str

    document: Any = field(repr=False)

    gdoc_body: List[GDoc_Content] = field(default=None)

    @classmethod
    def get_by_id(cls, document_id, auth=None, service=None, return_raw: bool = False):
        require_base_url = "https://docs.googleapis.com/"

        if not auth and not service:
            raise ga.GAuth_Required()

        if not service or service._baseUrl != require_base_url:
            if not auth:
                raise ga.GAuth_Required()

            service = auth.generate_service("docs", "v1")

        doc = get_document_by_id(service=service, document_id=document_id)

        gdoc_body = [
            GDoc_Content._from_json(content)
            for content in doc.get("body", {}).get("content")
        ]

        if return_raw:
            return doc

        return cls(
            auth=auth,
            service=service,
            document_id=document_id,
            document=doc,
            gdoc_body=gdoc_body,
        )

# %% ../../nbs/google/GDoc.ipynb 11
@dataclass
class DocState:
    cmds: List[dict] = field(default_factory=lambda: [])
    start_index: int = None
    end_index: int = None
    text: str = None

    function_ls: List[Callable] = field(default_factory=lambda: [])

    def run(self):
        [fn(self) for fn in self.function_ls]

        return self.cmds


def generate_delete_json(
    state: DocState, start_index: int = None, end_index: int = None
):
    start_index = start_index or state.start_index
    end_index = end_index or state.end_index

    state.cmds.append(
        {
            "deleteContentRange": {
                "range": {"startIndex": start_index, "endIndex": end_index}
            }
        }
    )
    state.start_index = start_index
    state.end_index = end_index


def generate_insert_json(
    state: DocState,
    start_index=None,
    text: str = None,
    is_suppress_add_new_line: bool = False,
):
    if not is_suppress_add_new_line and not text.endswith("\n"):
        text += "\n"

    start_index = start_index or state.start_index
    text = text or state.text

    state.cmds.append(
        {
            "insertText": {
                "location": {"index": start_index},
                "text": text,
            }
        }
    )

    state.start_index = start_index
    state.text = text
    state.end_index = len(text)


def generate_update_style(
    state: DocState,
    style_type: GDoc_StyleEnum = None,
    start_index=None,
    text=None,
):
    start_index = start_index or state.start_index
    text = text or state.text
    end_index = len(text) + start_index

    state.cmds.append(
        {
            "updateParagraphStyle": {
                "range": {"startIndex": start_index, "endIndex": end_index},
                "paragraphStyle": {"namedStyleType": style_type.value},
                "fields": "namedStyleType",
            }
        }
    )

    state.start_index = start_index
    state.text = text
    state.end_index = len(text)


def generate_insert_hyperlink(
    state: DocState,
    start_index: int = None,
    text: str = None,
    url: str = None,
):
    start_index = start_index or state.start_index
    text = text or state.text
    end_index = len(text) + start_index
    url = url or state.url

    state.cmds.append(
        {
            "updateTextStyle": {
                "textStyle": {
                    "link": {"url": url}  # Please set the modified URL here.
                },
                "range": {"startIndex": start_index, "endIndex": end_index},
                "fields": "link",
            }
        }
    )

    state.start_index = start_index
    state.text = text
    state.end_index = len(text)
    state.url = url
